/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user or with the express written consent of
 * Sun Microsystems, Inc.
 *
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 *
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 *
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 *
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
#if !defined(lint) && defined(SCCSIDS)
static char sccsid[] = "@(#)getrpcent.c 1.9 92/06/17 Copyr 1984 Sun Micro";
#endif

/*
 * Copyright (c) 1984 by Sun Microsystems, Inc.
 */


#include <stdio.h>
#include <sys/types.h>
#include <rpc/trace.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/file.h>
#include <dlfcn.h>
#include <limits.h>
#include <nsswitch.h>
#include <syslog.h>
#include <rpc/rpcent.h>

static struct translator {
	struct rpcent *(*rpname)();
	struct rpcent *(*rpnumber)();
	struct rpcent *(*rpgetent)();
	void *(*rpsetent)();
	void *(*rpendent)();
	char	module_name[_POSIX_PATH_MAX];
	void	*module_handle;
	struct translator	*next;
};

#define	NAME 1
#define	ADDR 2

static struct namaddr {
	int u_type;
	union {
		char *nam;
		int number;
	} u;
};

/* default policy is the same as in SunOS 4.1 */
static struct __nsw_lookup lkp2 = { "files", {1, 1, 1, 1}, NULL, NULL};
static struct __nsw_lookup lkp1 = { "nis", {1, 0, 0, 0}, NULL, &lkp2};
static struct __nsw_switchconfig rpc_default = { 0, "rpc", 2, &lkp1 };

static struct translator *xlate_list = NULL;
static struct __nsw_switchconfig *conf = NULL;
static int set_flag = 0;
static int stayopen;

/* the following two are directly set/used only by the *ent() routines */
static struct __nsw_lookup *cur_ns = NULL;
static struct translator *cur_xlate;

static struct rpcent *getrpcbynamadr();
static struct rpcent *callit();
static struct translator *load_xlate();

/*
 * This routine now acts as the user interface that goes through
 * the modules which implement the back-ends for each naming system in
 * the order specified by the configuration, until the name is
 * resolved, a hard error encountered or it runs out of all modules.
 */

struct rpcent *
getrpcbyname(name)
	char *name;
{
	struct namaddr namaddr;
	struct rpcent *dummy;

	trace1(TR_getrpcbyname, 0);
	namaddr.u_type = NAME;
	namaddr.u.nam = name;
	dummy = getrpcbynamadr(namaddr);
	trace1(TR_getrpcbyname, 1);
	return (dummy);
}

struct rpcent *
getrpcbynumber(number)
	int number;
{
	struct namaddr namaddr;
	struct rpcent *dummy;

	trace2(TR_getrpcbynumber, 0, number);
	namaddr.u_type = ADDR;
	namaddr.u.number = number;
	dummy = getrpcbynamadr(namaddr);
	trace1(TR_getrpcbynumber, 1);
	return (dummy);
}

static struct rpcent *
getrpcbynamadr(namaddr)
	struct namaddr namaddr;
{
	struct	translator   *xlate; /* pointer to translator list */
	char	*cur_module;
	struct	__nsw_lookup *lkp;
	struct	rpcent *rpcent = NULL;
	int	nserr;

	trace1(TR_getrpcbynamadr, 0);
	setrpcent(0);
	for (lkp = conf->lookups; lkp; lkp = lkp->next) {
		cur_module = lkp->service_name;
		nserr = __NSW_SUCCESS;
		xlate = load_xlate(&nserr, cur_module);
		if (!xlate && __NSW_ACTION(lkp, nserr) == __NSW_RETURN)
			break;
		if (!xlate)
			continue;
		nserr = __NSW_SUCCESS;
		rpcent = callit(xlate, namaddr, &nserr);
		if (__NSW_ACTION(lkp, nserr) == __NSW_RETURN)
			break;
	}
cleanup:
	/* __nsw_freeconfig(conf); free mem */
	endrpcent();
	trace1(TR_getrpcbynamadr, 1);
	return (rpcent);
}

setrpcent(stay)
	int stay;
{
	enum __nsw_parse_err pserr;
	int	nserr = __NSW_SUCCESS;
	struct translator *xlate;

	trace2(TR_setrpcent, 0, stay);
	stayopen |= stay;
	if (!conf &&
		(conf = __nsw_getconfig(__NSW_RPC_DB, &pserr)) == NULL)
		conf = &rpc_default;
	/*
	 * if getrpcent was the previous call and it was in the
	 * naming service other than the "first" one, "endrpcent" it
	 * and reset the current naming service to the first one.
	 */
	if (cur_ns && (cur_ns != conf->lookups)) {
		xlate = load_xlate(&nserr, cur_ns->service_name);
		if (xlate && xlate->rpendent)
			(*(xlate->rpendent))(&nserr);
	}
	/*
	 * set the first naming service as the current one, and
	 * load its op vector.
	 */
	cur_ns = conf->lookups;
	if (cur_ns) {
		cur_xlate = load_xlate(&nserr, cur_ns->service_name);
		if (cur_xlate && cur_xlate->rpsetent)
			(*(cur_xlate->rpsetent))(&nserr, stayopen);
	} else
		cur_xlate = NULL;
	set_flag = 1;
	trace2(TR_setrpcent, 1, stay);
	return (0);
}

struct rpcent *
getrpcent()
{
	struct rpcent *rp = NULL;
	int nserr;

	trace1(TR_getrpcent, 0);
	if (!set_flag)
		setrpcent(0);
repeat:
	nserr = __NSW_SUCCESS;
	if (!cur_xlate && cur_ns)
		cur_xlate = load_xlate(&nserr, cur_ns->service_name);
	/*
	 * success status returned by load_xlate is not the same as the
	 * success specified by the switch, hence take action only on
	 * non-success error codes from load_xlate.
	 */
	if (!cur_xlate &&
		((__NSW_ACTION(cur_ns, nserr) == __NSW_RETURN) ||
		(cur_ns->next == NULL))) {
		trace1(TR_getrpcent, 1);
		return (NULL);
	}
	if (!cur_xlate ||
		(!cur_xlate->rpgetent &&
		    __NSW_ACTION(cur_ns, __NSW_UNAVAIL) == __NSW_CONTINUE)) {
		cur_ns = cur_ns->next;
		nserr = __NSW_SUCCESS;
		cur_xlate = load_xlate(&nserr, cur_ns->service_name);
		goto repeat;
	}
	nserr = __NSW_SUCCESS;
	if (cur_xlate->rpgetent)
		rp = (*(cur_xlate->rpgetent))(&nserr);
	else
		nserr = __NSW_UNAVAIL;
	if (__NSW_ACTION(cur_ns, nserr) == __NSW_RETURN) {
		trace1(TR_getrpcent, 1);
		return (rp);
	}
	/*
	 * The getent failed and the action points to the next naming service,
	 * or the policy setter is a weirdo and asked to continue on
	 * success. Again go to the next naming service.
	 */
	if (cur_xlate->rpendent)
		(*(cur_xlate->rpendent))(&nserr);
	if (cur_ns->next == NULL) {
		trace1(TR_getrpcent, 1);
		return (NULL);
	}
	cur_ns = cur_ns->next;
	cur_xlate = load_xlate(&nserr, cur_ns->service_name);
	if (cur_xlate && cur_xlate->rpsetent)
		(*(cur_xlate->rpsetent))(&nserr, stayopen);
	goto repeat;
}

void
endrpcent()
{
	int nserr;

	trace1(TR_endrpcent, 0);
	if (cur_xlate && cur_xlate->rpendent)
		(*(cur_xlate->rpendent))(&nserr);
	cur_ns = NULL;
	cur_xlate = NULL;
	set_flag = 0;
	stayopen = 0;
	trace1(TR_endrpcent, 1);
}

static struct translator *
load_xlate(nserr, module_name)
	int *nserr;
	char *module_name;
{
	struct  translator   *xlate;
	char    full_mod_name[_POSIX_PATH_MAX];
	char    *mp = full_mod_name;
	void    *mod_handle;

	trace1(TR_load_xlate, 0);
	for (xlate = xlate_list; xlate; xlate = xlate->next)
		if (strcmp(module_name, xlate->module_name) == 0) {
			trace1(TR_load_xlate, 1);
			return (xlate);
		}
	memset(mp, 0, _POSIX_PATH_MAX);
	sprintf(mp, "%s%s%s", __NSW_LIB, module_name, ".so");
	mod_handle = dlopen(mp, RTLD_LAZY);
	if (mod_handle == NULL) {
		*nserr = __NSW_UNAVAIL;
		trace1(TR_load_xlate, 1);
		return (NULL);
	}
	xlate = (struct translator *)malloc(sizeof (struct translator));
	if (xlate == NULL) {
		*nserr = __NSW_UNAVAIL; /* XXX: really means something else ! */
		trace1(TR_load_xlate, 1);
		return (NULL);
	}
	xlate->next = xlate_list;
	xlate_list = xlate;
	xlate->module_handle = mod_handle;
	strcpy(xlate->module_name, module_name);
	xlate->rpname = (struct rpcent *(*)())dlsym(mod_handle,
							"_getrpcbyname");
	xlate->rpnumber = (struct rpcent *(*)())dlsym(mod_handle,
							"_getrpcbynumber");
	xlate->rpgetent = (struct rpcent *(*)())dlsym(mod_handle, "_getrpcent");
	xlate->rpsetent = (void *(*)())dlsym(mod_handle, "_setrpcent");
	xlate->rpendent = (void *(*)())dlsym(mod_handle, "_endrpcent");
	*nserr = __NSW_SUCCESS;
	trace1(TR_load_xlate, 1);
	return (xlate);
}

/*
 * Do the actual work of calling the appropriate routine by
 * switching on namaddr, return the rpcent if successful;
 * otherwise NULL and error.
 */

static struct rpcent *
callit(xlate, namaddr, errp)
struct translator *xlate;
struct namaddr namaddr;
int *errp;
{
	struct rpcent *rp = NULL;
#ifdef DEBUG
	char *dl;
#endif

	trace1(TR_callit, 0);
	*errp = __NSW_SUCCESS;
	switch (namaddr.u_type) {
	case NAME:
#ifdef DEBUG
		syslog(LOG_DEBUG, "getrpcbyname: in %s for %s\t",
				xlate->module_name, namaddr.u.nam);
#endif
		if (!(xlate->rpname))
			*errp = __NSW_UNAVAIL;
		else
			rp = (*(xlate->rpname))(errp, namaddr.u.nam);
#ifdef DEBUG
		if (rp)
			syslog(LOG_DEBUG, "succeeds\n");
		else
			syslog(LOG_DEBUG, "fails !\n");
#endif
		break;
	case ADDR:
#ifdef DEBUG
		syslog(LOG_DEBUG, "getrpcbynumber: in %s for %d\t",
				xlate->module_name, namaddr.u.number);
#endif
		if (!(xlate->rpnumber))
			*errp = __NSW_UNAVAIL;
		else
			rp = (*(xlate->rpnumber))(errp, namaddr.u.number);
#ifdef DEBUG
		if (rp)
			syslog(LOG_DEBUG, "succeeds\n");
		else
			syslog(LOG_DEBUG, "fails !\n");
#endif
		break;
	default:
		break;
	}
#ifdef DEBUG
	dl = dlerror();
	if (dl)
		syslog(LOG_ALERT, "gethost: backend %s\n", dl);
#endif
	trace1(TR_callit, 1);
	return (rp);
}
